/*
    Copyright 2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
    #define usleep  sleep
    #define ONEMS   1
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define ONEMS   1000
#endif



#define VER         "0.1a"
#define PORT        135
#define BUFFSZ      1500        // avoids fragments actually not supported
#define CALCLEN     (BUFFSZ                       - ( \
                    sizeof(RPCDCE)                  + \
                    sizeof(NETR) + servermsglen + 3 + \
                    sizeof(NETR) + clientmsglen + 3 + \
                    sizeof(NETR)))



#pragma pack(1)

typedef struct {
    u_int   g1;
    u_short g2;
    u_short g3;
    u_char  g4;
    u_char  g5;
    u_char  g6;
    u_char  g7;
    u_char  g8;
    u_char  g9;
    u_char  g10;
    u_char  g11;
} guid;

typedef struct {
    u_char  version;
    u_char  packet_type;
    u_char  flags1;
    u_char  flags2;
    u_char  data_representation[3];
    u_char  serial_high;
    guid    object_uuid;
    guid    iface;
    guid    activity;
    u_int   boot_time;
    u_int   iface_ver;
    u_int   sequence_num;
    u_short opnum;
    u_short iface_hint;
    u_short activity_hint;
    u_short fragment_len;
    u_short fragment_num;
    u_char  auth_proto;
    u_char  serial_low;
} RPCDCE;

typedef struct {
    u_int   max_count;
    u_int   offset;
    u_int   actual_count;
    u_char  message[];
} NETR;

#pragma pack()



void *add_entry(NETR *entry, u_char *data, int len);
u_char *load_file(u_char *fname, int *datalen);
int input_message(u_char *msg, int size);
void create_rand_data(u_char *data, int len, u_int rnd);
void setguid(u_char *myguid, guid *dst);
u_int resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    RPCDCE  *rpcdce;
    NETR    *server,
            *client,
            *message;
    int     sd,
            len,
            i,
            servermsglen,
            clientmsglen,
            datalen,
            flooding   = -1;
    u_short port       = PORT;
    u_char  *buff,
            *servermsg = "",
            *clientmsg = "",
            *data      = NULL;
    void    *next;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "DCE Messenger "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s [options] <host>\n"
            "\n"
            "Options:\n"
            "-f FILE   gets the message from the file FILE\n"
            "-m \"MSG\"  one-line message\n"
            "-s SRC    source name (none)\n"
            "-d DST    destination name (none)\n"
            "-x MS     flooding, will be sent a message each MS milliseconds\n"
            "-p PORT   target port (%hu), this option is probably useless...\n"
            "\n",
            argv[0], port);
        exit(1);
    }

    argc--;
    for(i = 1; i < argc; i++) {
        switch(argv[i][1]) {
            case 'f': data      = load_file(argv[++i], &datalen);   break;
            case 'm': datalen   = strlen(data = argv[++i]);         break;
            case 's': servermsg = argv[++i];                        break;
            case 'd': clientmsg = argv[++i];                        break;
            case 'x': flooding  = atoi(argv[++i]);                  break;
            case 'p': port      = atoi(argv[++i]);                  break;
            default: {
                printf("\nError: wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
                } break;
        }
    }

    peer.sin_addr.s_addr = resolv(argv[argc]);
    peer.sin_port        = htons(port);
    peer.sin_family      = AF_INET;

    printf("- target        %s : %hu\n",
        inet_ntoa(peer.sin_addr), port);

    servermsglen = strlen(servermsg);
    clientmsglen = strlen(clientmsg);

    if(data) {
        if(datalen > CALCLEN) datalen = CALCLEN;
    } else {
        datalen = CALCLEN;
        data = malloc(datalen + 1);
        if(!data) std_err();
        datalen = input_message(data, datalen);
    }

    buff = malloc(BUFFSZ);
    if(!buff) std_err();

    rpcdce                  = (void *)buff;
    rpcdce->version         = 4;
    rpcdce->packet_type     = 0;
    rpcdce->flags1          = 0x78;
    rpcdce->flags2          = 0;
    memcpy(rpcdce->data_representation, "\x10\x00\x00", 3);
    rpcdce->serial_high     = 0;
    setguid("00000000-0000-0000-0000-000000000000", &rpcdce->object_uuid);
    setguid("5a7b91f8-ff00-11d0-a9b2-00c04fb6e6fc", &rpcdce->iface);
    create_rand_data((void *)&rpcdce->activity, sizeof(guid), time(NULL));
    rpcdce->boot_time       = 0;
    rpcdce->iface_ver       = 1;
    rpcdce->sequence_num    = 0;
    rpcdce->opnum           = 0;
    rpcdce->iface_hint      = 0xffff;
    rpcdce->activity_hint   = 0xffff;
    rpcdce->fragment_num    = 0;
    rpcdce->auth_proto      = 0;
    rpcdce->serial_low      = 0;

    server                  = (void *)((void *)rpcdce + sizeof(RPCDCE));
    next                    = add_entry(server, servermsg, servermsglen);

    client                  = next;
    next                    = add_entry(client, clientmsg, clientmsglen);

    message                 = next;
    next                    = add_entry(message, data, datalen);

    len                     = (u_char *)next - buff;
    rpcdce->fragment_len    = len - sizeof(RPCDCE);

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

sendmsg:
    if(sendto(sd, buff, len, 0, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();

    if(flooding >= 0) {
        fputc('.', stdout);
        usleep(flooding * ONEMS);
        goto sendmsg;
    }

    close(sd);
    printf("- message sent: %u bytes / packet %u bytes\n\n",
        datalen, len);
    return(0);
}



void *add_entry(NETR *entry, u_char *data, int len) {
    int     pad;

    entry->max_count       = len + 1;   // +1 for NULL byte REQUIRED!!!
    entry->offset          = 0;
    entry->actual_count    = entry->max_count;
    memcpy(entry->message, data, len);

    pad = entry->max_count & 3;
    if(pad) pad = 4 - pad;

    memset((void *)entry->message + len, 0, (entry->max_count + pad) - len);
    return((void *)entry->message + entry->max_count + pad);
}



u_char *load_file(u_char *fname, int *datalen) {
    struct  stat    xstat;
    FILE    *fd;
    u_char  *data;

    printf("- load file:    %s\n", fname);
    fd = fopen(fname, "rb");
    if(!fd) std_err();
    fstat(fileno(fd), &xstat);
    *datalen = xstat.st_size;
    data = malloc(*datalen);
    if(!data) std_err();
    fread(data, *datalen, 1, fd);
    fclose(fd);
    return(data);
}



int input_message(u_char *msg, int size) {
    int     msglen;
    u_char  *p,
            *l;

    p      = msg;
    msglen = size;

    fputs("- type your message and then put a single dot (.) for sending it:\n", stdout);
    fflush(stdin);
    while(fgets(p, msglen, stdin)) {
        for(l = p; *l && (*l != '\n') && (*l != '\r'); l++);
        *l = 0;
        if(!strcmp(p, ".")) break;
        p = l;
        *p++ = '\r';
        *p++ = '\n';
        msglen = size - (p - msg);
    }
    *p = 0;

    return(p - msg);
}



void create_rand_data(u_char *data, int len, u_int rnd) {
    while(len--) {
        rnd = (rnd * 0x343FD) + 0x269EC3;
        *data++ = rnd;
    }
}



void setguid(u_char *myguid, guid *dst) {
    int     g1,g2,g3,g4,g5,g6,g7,g8,g9,g10,g11;

    sscanf(
        myguid,
        "%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
        &g1,
        &g2,
        &g3,
        &g4,
        &g5,
        &g6,
        &g7,
        &g8,
        &g9,
        &g10,
        &g11);

    dst->g1 = g1;
    dst->g2 = g2;
    dst->g3 = g3;
    dst->g4 = g4;
    dst->g5 = g5;
    dst->g6 = g6;
    dst->g7 = g7;
    dst->g8 = g8;
    dst->g9 = g9;
    dst->g10 = g10;
    dst->g11 = g11;
}



u_int resolv(char *host) {
    struct hostent *hp;
    u_int  host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif



